/**********************************************************************
 
	Copyright (C) 2003 Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/



#include	<stdlib.h>
#include	"memory_debug.h"
#include	"xlerror.h"
#include	"memory_routine.h"
#include	"utils.h"
#include	"xl.h"
#include	"resource.h"
#include	"lock_level.h"
#include	"task.h"
#include	"queue.h"
#include	"pri_level.h"

int default_dgb_pixel = DEFAULT_DGB_PIXEL;
SEM draw_lock;
void (*draw_gb_indicate_task)(TKEY);
void gc_gb_sexp(XL_SEXP*);

void
init_vector()
{
	draw_lock = new_lock(LL_DRAW);

}

void
pdb_err_sensitive(RESOURCE * r,int on_flag)
{
	lock_task(draw_lock);
	r->draw_gb.pdb_err_sensitive = on_flag;
	wakeup_task((int)&r->draw_gb.pdb_lock);
	unlock_task(draw_lock,"lock_pdb_lock");
}


int
test_pdb_err_sensitive(RESOURCE * r)
{
int ret;
	lock_task(draw_lock);
	if ( r->draw_gb.pdb_err_sensitive )
		ret = 1;
	else	ret = 0;
	unlock_task(draw_lock,"lock_pdb_err_sensitive");
	return ret;
}

int
xx_lock_pdb_lock(RESOURCE * r,int err_sensitive,char *file,int line)
{
int t;
int pri;
	if ( r == 0 )
		return -2;
	pri = push_pri(PRI_LOCK);
	lock_task(draw_lock);
	if ( r->h.type != RT_DRAW_GB ) {
		unlock_task(draw_lock,"lock_pdb_lock");
		change_pri(0,pri);
		return -2;
	}
	t = get_tid();
	if ( r->draw_gb.pdb_lock == 0 )
		goto next;
	if ( r->draw_gb.pdb_lock_task == t )
		goto next;
	for ( ; r->draw_gb.pdb_lock ; ) {
		if ( err_sensitive && r->draw_gb.pdb_err_sensitive ) {
			unlock_task(draw_lock,"lock_pdb_lock");
			change_pri(0,pri);
			return -1;
		}
		_i_sleep_task((int)&r->draw_gb.pdb_lock,draw_lock,file,line);
		lock_task(draw_lock);
	}
next:
	r->draw_gb.pdb_lock_time = get_xltime();
	r->draw_gb.pdb_lock_file = file;
	r->draw_gb.pdb_lock_line = line;
	r->draw_gb.pdb_lock ++;
	r->draw_gb.pdb_lock_task = t;
	r->draw_gb.pdb_lock_pri = pri;
	unlock_task(draw_lock,"lock_pdb_lock");
	return 0;
}


void
unlock_pdb_lock(RESOURCE * r)
{
int pri;
	lock_task(draw_lock);
	if ( r->draw_gb.pdb_lock ) {
		if ( r->draw_gb.pdb_lock_task == 0 )
			goto next;
		if ( r->draw_gb.pdb_lock_task != get_tid() )
			er_panic("unlock_pdb_lock");
		r->draw_gb.pdb_lock --;
	next:
		if ( r->draw_gb.pdb_lock == 0 ) {
			r->draw_gb.pdb_lock_task = 0;
			wakeup_task((int)&r->draw_gb.pdb_lock);
		}
	}
	pri = r->draw_gb.pdb_lock_pri;
	unlock_task(draw_lock,"unlock_pdb_lock");
	change_pri(0,pri);
/*
printf("unlock_pdb 1 %ls %i\n",
get_url_str2(&r->h.entry),
get_tid());
*/
}

int
gc_gv_draw_func(AVT_NODE * a)
{
OBJ * o;
	o = a->data;
	gc_gb_sexp(o->h.info_org);
	gc_gb_sexp(o->h.info_card);
	gc_gb_sexp(o->h.info_onmap);
	return 0;
}

void
gc_gv_draw(RESOURCE * r)
{
  //char * url = n_string(std_cm,get_url_str2(&r->h.entry));
	avt_trace_from_small(r->draw_gb.obj_code_tree,
		gc_gv_draw_func,0);
}


XL_SEXP *
gv_new_vector(RESOURCE * r,XL_SEXP * s,int ds)
{
int i;
int ret_dpm,ret_level;
double dpm;
int level;
L_CHAR * l_unit,* d_unit;
char * e_param;
XL_SEXP * e_data;
int er;
double pitch_x,pitch_y;
double _x,_y;
L_CHAR * pitch_unit_x,* pitch_unit_y;
int ret_pitch_x,ret_pitch_y;
REAL1 _x_len,_y_len,__x_len,__y_len;

	if ( ds == NR_KEEP )
		return 0;
	if ( r->h.target.resource == 0 )
		goto non_prog;
	i = l_strlen(r->h.target.resource);
	i--;
	for ( ; i >= 0 ; i -- )
		if ( r->h.target.resource[i] == '.' )
			break;
	if ( r->h.target.resource[i] != '.' )
		goto non_prog;
	if ( l_strcmp(&r->h.target.resource[i],l_string(std_cm,".til")) == 0 )
		goto pdb_tile;
	if ( l_strcmp(&r->h.target.resource[i],
			l_string(std_cm,".pp")) == 0 )
		goto pdb_prog;
non_prog:


	r->draw_gb.flags = 0;
	r->draw_gb.type = DGT_LOAD;
	r->draw_gb.level = 0;
	r->draw_gb.resolution = 0;
	r->draw_gb.hit_level = 0;
	r->draw_gb.hit_outofcircle = 0;

	new_vector_option1(r,ds);

	memset(&r->draw_gb.ir_que,0,sizeof(SYS_QUEUE));
	if ( draw_gb_indicate_task ) {
		r->draw_gb.ir_que.flags = QF_STACK;
		r->draw_gb.ir_que.gc_func = 0;
		r->draw_gb.ir_que.gc_get = 0;
		r->draw_gb.ir_que.key_func = draw_gb_indicate_task;
		r->draw_gb.ir_que.pri = PRI_FETCH;
		r->draw_gb.ir_que.work = r;
		setup_queue(&r->draw_gb.ir_que);
	}
	r->draw_gb.ischeme = 0;
	r->draw_gb.onmap_indicate_size = DEFAULT_ONMAP_INDICATE_SIZE;
	return 0;
pdb_prog:


	r->draw_gb.type = DGT_PDB;
	r->draw_gb.flags = 0;
	r->draw_gb.hit_level = 0;
	r->draw_gb.hit_outofcircle = 0;
	get_field(s,
		l_string(std_cm,"level"),"i",&level,&ret_level,&l_unit,
		l_string(std_cm,"resolution"),"lf",&dpm,&ret_dpm,&d_unit,
		0);
	if ( ret_dpm || ret_level ) {
		e_param = "resolution or level error";
		e_data = s;
		goto subtype_error;
	}
	r->h.loop_no = 0;
	r->draw_gb.tile_id = 0;
	r->draw_gb.level = level;
	r->draw_gb.resolution = conv_unit(&er,
		r->h.cu.uenv,
		dpm,d_unit,
		reso_c_unit(&r->h.cu));
	r->draw_gb.matrix_ofs.x =
		(1+DM_RATE)*r->h.minrect.tl.x
		- DM_RATE*r->h.minrect.br.x;
	r->draw_gb.matrix_ofs.y =
		(1+DM_RATE)*r->h.minrect.tl.y
		- DM_RATE*r->h.minrect.br.y;
	if ( ds == NR_CLEAR )
		d_f_ree(r->draw_gb.pitch_list);
	r->draw_gb.pitch_list
		= d_alloc(sizeof(GB_POINT)*level);
	r->draw_gb.pitch_list[level-1].x =
		(r->h.minrect.br.x - r->h.minrect.tl.x)*(1+2*DM_RATE)/2;
	r->draw_gb.pitch_list[level-1].y =
		(r->h.minrect.br.y - r->h.minrect.tl.y)*(1+2*DM_RATE)/2;
	for ( i = level-2 ; i >= 0 ; i -- ) {
		r->draw_gb.pitch_list[i].x
			= r->draw_gb.pitch_list[i+1].x/DM_DIM;
		r->draw_gb.pitch_list[i].y
			= r->draw_gb.pitch_list[i+1].y/DM_DIM;
	}
	if ( r->draw_gb.pitch_list[0].x <
			default_dgb_pixel/r->draw_gb.resolution ) {
		r->draw_gb.pitch_list[0].x =
			default_dgb_pixel/r->draw_gb.resolution;
		for ( i = 1 ; i < level ; i ++ )
			r->draw_gb.pitch_list[i].x
				= r->draw_gb.pitch_list[i-1].x*DM_DIM;
	}
	if ( r->draw_gb.pitch_list[0].y <
			default_dgb_pixel/r->draw_gb.resolution ) {
		r->draw_gb.pitch_list[0].y =
			default_dgb_pixel/r->draw_gb.resolution;
		for ( i = 1 ; i < level ; i ++ )
			r->draw_gb.pitch_list[i].y
				= r->draw_gb.pitch_list[i-1].y*DM_DIM;
	}
	new_vector_option2(r,ds,level);

	memset(&r->draw_gb.ir_que,0,sizeof(SYS_QUEUE));
	if ( draw_gb_indicate_task ) {
		r->draw_gb.ir_que.flags = QF_STACK;
		r->draw_gb.ir_que.gc_func = 0;
		r->draw_gb.ir_que.gc_get = 0;
		r->draw_gb.ir_que.key_func = draw_gb_indicate_task;
		r->draw_gb.ir_que.pri = PRI_FETCH;
		r->draw_gb.ir_que.work = r;
		setup_queue(&r->draw_gb.ir_que);
	}

	r->draw_gb.ischeme = 0;
	r->draw_gb.onmap_indicate_size = DEFAULT_ONMAP_INDICATE_SIZE;
	return 0;
pdb_tile:


	r->draw_gb.type = DGT_TILE;
	r->draw_gb.tile_id = 0;
	r->draw_gb.flags = DGF_INFO|DGF_MARK;
	r->draw_gb.hit_level = 0;
	r->draw_gb.hit_outofcircle = 0;
	get_field(s,
		l_string(std_cm,"pitch-x"),"lf",&pitch_x,&ret_pitch_x,&pitch_unit_x,
		l_string(std_cm,"pitch-y"),"lf",&pitch_y,&ret_pitch_y,&pitch_unit_y,
		l_string(std_cm,"resolution"),"lf",&dpm,&ret_dpm,&d_unit,
		0);
	pitch_x = conv_unit(&er,
			r->h.cu.uenv,
			pitch_x,
			pitch_unit_x,
			r->h.cu.unit);
	pitch_y = conv_unit(&er,
			r->h.cu.uenv,
			pitch_y,
			pitch_unit_y,
			r->h.cu.unit);
	if ( ret_dpm ) {
		e_param = "resolution or level error";
		e_data = s;
		goto subtype_error;
	}
	_x = pitch_x;
	_y = pitch_y;
	_x_len = __x_len = r->h.minrect.br.x - r->h.minrect.tl.x;
	_y_len = __y_len = r->h.minrect.br.y - r->h.minrect.tl.y;
	for ( level = 0 ;
		_x < _x_len ||
		_y < _y_len;
		level ++ , 
		_x = _x*DM_DIM , 
		_y = _y*DM_DIM );
	level ++;

	r->h.loop_no = 0;
	r->draw_gb.level = level;
	r->draw_gb.resolution = conv_unit(&er,
		r->h.cu.uenv,
		dpm,d_unit,
		reso_c_unit(&r->h.cu));
	r->draw_gb.matrix_ofs.x =
		(r->h.minrect.tl.x + r->h.minrect.br.x - _x)/2;
	r->draw_gb.matrix_ofs.y =
		(r->h.minrect.tl.y + r->h.minrect.br.y - _y)/2;
	if ( ds == NR_CLEAR )
		d_f_ree(r->draw_gb.pitch_list);
	r->draw_gb.pitch_list
		= d_alloc(sizeof(GB_POINT)*level);
	r->draw_gb.pitch_list[level-1].x = _x;
	r->draw_gb.pitch_list[level-1].y = _y;
	for ( i = level-2 ; i >= 0 ; i -- ) {
		r->draw_gb.pitch_list[i].x
			= r->draw_gb.pitch_list[i+1].x/DM_DIM;
		r->draw_gb.pitch_list[i].y
			= r->draw_gb.pitch_list[i+1].y/DM_DIM;
	}
	if ( r->draw_gb.pitch_list[0].x <
			default_dgb_pixel/r->draw_gb.resolution ) {
		log_printf(LOG_ERROR,LOG_LAYER_GB,0,
			"TILE VECTOR pitch.x = %f resolution= %f\n",
			r->draw_gb.pitch_list[0].x,
			default_dgb_pixel/r->draw_gb.resolution);
	}
	if ( r->draw_gb.pitch_list[0].y <
			default_dgb_pixel/r->draw_gb.resolution ) {
		log_printf(LOG_ERROR,LOG_LAYER_GB,0,
			"TILE VECTOR pitch.x = %f resolution= %f\n",
			r->draw_gb.pitch_list[0].y,
			default_dgb_pixel/r->draw_gb.resolution);
	}
	new_vector_option2(r,ds,level);

	memset(&r->draw_gb.ir_que,0,sizeof(SYS_QUEUE));
	if ( draw_gb_indicate_task ) {
		r->draw_gb.ir_que.flags = QF_STACK;
		r->draw_gb.ir_que.gc_func = 0;
		r->draw_gb.ir_que.gc_get = 0;
		r->draw_gb.ir_que.key_func = draw_gb_indicate_task;
		r->draw_gb.ir_que.pri = PRI_FETCH;
		r->draw_gb.ir_que.work = r;
		setup_queue(&r->draw_gb.ir_que);
	}

	r->draw_gb.ischeme = 0;
	r->draw_gb.onmap_indicate_size = DEFAULT_ONMAP_INDICATE_SIZE;
	return 0;
subtype_error:
	return get_error(
		s->h.file,
		s->h.line,
		XLE_PROTO_INV_PARAM,
		l_string(std_cm,"gv-new-resource(draw_gb)"),
		List(n_get_string("invalid subtype in the meta info"),
			get_string(l_string(std_cm,e_param)),
			n_get_string("argument"),
			e_data,
			-1));
}

