/**********************************************************************
 
	Copyright (C) 2003 Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/


#include	"win_flame.h"
#include	"memory_debug.h"



int
e2d_p_exit(GBVIEW_FLAME * gf,WIN_FLAME * wf)
{
WIN_FLAME ** wfp;


	if ( !(wf->flags & WFF_FREE) ) {
		return 0;
	}
	if ( gf->flame_base == wf ) {
		if ( _wf_set_current(gf,wf,0) )
			assert_ls(gf,LSF_CURRENT);
		else {
			gf->flame_base = 0;
		}
	}
	free_mh(wf->mh);
	wf_free_si(wf);
	wf_close_work_area_set(wf->wa_set);
	for ( wfp = &gf->flame ; *wfp && *wfp != wf ; wfp = &(*wfp)->next );
	if ( *wfp )
		*wfp = wf->next;
	wakeup_task((int)wf);
	wf_free_lock_list(wf);
	d_f_ree(wf);
	_wf_flame_redraw(gf);
	call_gv_event(RID_INDICATED_COORD,l_string(std_cm,"hide"));
	call_gv_event(RID_HIDE_COORD,l_string(std_cm,"hide"));
	return 1;
}


int
e2d_p_move(
	GBVIEW_FLAME * gf,
	WIN_FLAME * wf,VPOINT from,VPOINT to,void * _ix)
{
VPOINT * ix;
int dx,dy;
	ix = _ix;
	wf_move_si(gf, wf, dx = to.x - from.x,dy =  to.y - from.y);
	e2d_move_wa(wf->wa_set,dx,dy);
	wf->flags |= WFF_PLOT_DIRTY;
	return 0;
}


int
e2d_p_change(
	WIN_FLAME_TABLE * tbl,
	GBVIEW_FLAME * gf,
	WIN_FLAME * wf)
{
GV_RESOURCE_LIST * rl;
	switch ( wf->tbl->wf_type ) {
	case WFT_STD:
		d_f_ree(wf->pixels);
		wf->pixels = 0;
		break;
	case WFT_SMALL:
		d_f_ree(wf->pixels);
		wf->pixels = 0;
		break;
	case WFT_PLOT:
		break;
	default:
		er_panic("e2d_change not support");
	}
	wf->tbl = tbl;
	rl = search_resource_list(gf,wf->draw,0,1);
	if ( rl == 0 )
		return 0;
	rl->weight = W_E2D_PLOT;
	return 0;
}



int
_e2d_p_redraw(GBVIEW_FLAME * gf,WIN_FLAME * wf)
{
int w,h;
MAP_HISTORY * mh;
WIN_FLAME * _wf;
SURP_SET surp;
int id;
GB_RECT r;
int ci_ret;

	r.tl.x = r.tl.y = 0;
	r.br.x = gf->win_width;
	r.br.y = gf->win_height;

	wf_lock(gf);


	if ( wf->flags & WFF_RESIZE ) {
		wf_unlock(gf);
		return -2;
	}
	wf->mov_x = 0;
	wf->mov_y = 0;
	wf->mov_flag = 0;

	w = gf->win_width;
	h = gf->win_height;

	mh = add_mh(&gf->flame_base_display_map,wf->mh);
	set_surp(&surp,&r,gf->flame_base_resolution);
	map_from_flame(mh,surp.ptr,surp.reso,SURP_NOS);
	id = wf->id;
	wf_unlock(gf);

retry:
	ci_ret = cache_image(gf,wf->id,wf->draw,0,0,0,w*h,&r,mh,&surp,1,
		WFF_PLOT_DIRTY,0);

	if ( ci_ret & WFF_FLUSH ) {
	GB_RECT rct;
		wf_work_area_flush(wf->wa_set);
		rct.tl.x = 0;
		rct.br.x = gf->win_width;
		rct.tl.y = 0;
		rct.br.y = gf->win_height;
		win_flame_dirty(gf,wf->id,&rct,WFF_LUSTER_DIRTY,0);

		goto retry;
	}
		
	wf_lock(gf);
	_wf = _get_wf_ptr(gf,id);
	if ( _wf != wf )
		goto end;

	if ( wf->flags & WFF_RESIZE ) {
		free_mh(mh);
		wf_unlock(gf);
		return -2;
	}
	if ( wf->mov_flag )
		wf_free_si(wf);
	else	regulate_si(wf);

end:
	wf_unlock(gf);
	free_mh(mh);
	return 0;
}

int
e2d_p_redraw(GBVIEW_FLAME * gf,WIN_FLAME * wf)
{
int ret;
GV_RESOURCE_LIST * rl;
	ret = 0;
	wf_lock(gf);
	rl = _search_resource_list(gf,wf->draw,0,1);
	if ( rl == 0 ) {
		wf_unlock(gf);
		return 1;
	}
	rl->weight = W_E2D_PLOT;
	if ( wf->flags & WFF_PLOT_DIRTY ) {
		wf->flags &= ~(WFF_PLOT_DIRTY|WFF_RESIZE);
		wf_unlock(gf);
		_e2d_p_redraw(gf,wf);
		ret = 1;
	}
	else	wf_unlock(gf);
	return ret;
}

int
e2d_p_dirty(GBVIEW_FLAME * gf,WIN_FLAME * wf,GB_RECT * r,int flags,int mode)
{
	if ( r ) {
		if ( r->tl.x > r->br.x ||
				r->tl.y > r->br.y )
			goto ok;
		if ( r->tl.x >= gf->win_width )
			return 0;
		if ( r->br.x <= 0 )
			return 0;
		if ( r->tl.y >= gf->win_height )
			return 0;
		if ( r->br.y <= 0 )
			return 0;
	}

ok:

	wf->flags |= flags & WFF_DIRTY &
			wf_get_dirty_mask(wf->flags);
	_wakeup_wf(gf,wf);
	return 0;
}



int
e2d_p_zoom_etc(
	GBVIEW_FLAME * gf,
	WIN_FLAME * wf,...)
{

	wf_free_si(wf);
	e2d_zoom_wa(wf->wa_set);
	wf->flags |= WFF_PLOT_DIRTY;
	return 0;
}

int
e2d_p_resize(GBVIEW_FLAME * gf,WIN_FLAME * wf,int old_w,int old_h)
{
	e2d_resize_wa(wf->wa_set,old_w,old_h,gf->win_width,gf->win_height);
	return 0;
}


