/**********************************************************************
 
	Copyright (C) 2003 Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/

/********************************************

	memory manage and gc routine
	(c) H.Mori 1992.9.
	memory_routine.c


*********************************************/

//#define SPEED_DEBUG

#include "memory_debug.h"

#include <stdlib.h>
#include <stdio.h>
#include	"lock_level.h"
#include	"task.h"
#include "memory_debug.h"
#include "memory_routine.h"
#include "utils.h"

char * gc_file;
int gc_line;

void mem_test(char * msg);

#define GC_THRE  2000000
int GC_RESET = 50;
int GC_LOW   = 50;


SEM mem_lock;

#define INSERT_MEM(n,e)		\
	(e)->m_prev = (n);		\
	(e)->m_next = (n)->m_next;	\
	(n)->m_next = (e);		\
	(e)->m_next->m_prev = (e);
#define DELETE_MEM(e)	\
	(e)->m_prev->m_next = (e)->m_next;	\
	(e)->m_next->m_prev = (e)->m_prev;
#define INIT_MEM(n)	\
	(n)->m_next = (n)->m_prev = (n);

#define SIZEOF_MEM	sizeof(MEM)


void * dumy_mem;
int malloc_count;
int malloc_max = GC_THRE;
int abs_malloc;

MEM mem_header;
int mem_tid,mem_cnt;

#ifdef SPEED_DEBUG
#define S_MALLOC	s_alloc
#define S_FREE		s_free
#else
#define S_MALLOC	malloc
#define S_FREE		free
#endif


void
minit()
{
	mem_lock = new_lock(LL_MEMORY);
	INIT_MEM(&mem_header);
	mem_header.m_flags = 0;
}


void
lock_mem()
{
int tid;
	tid = get_tid();
	if ( mem_tid == tid ) {
		mem_cnt ++;
		return;
	}
	lock_task(mem_lock);
	mem_tid = tid;
	mem_cnt = 1;
}

void
unlock_mem()
{
int tid;
	tid = get_tid();
	if ( mem_tid != tid )
		er_panic("unlock_mem");
	mem_cnt --;
	if ( mem_cnt == 0 ) {
		mem_tid = 0;
		unlock_task(mem_lock,"unlock_mem");
	}
}

void
_mem_debug1()
{
void * a;
	a = S_MALLOC(56);
	S_FREE(a);
}


void *
xx_mmalloc(unsigned long size,void (*gc_func)(),char * __file,int __line)
{
void * ret;
int i;
char * pt;
MEM * m;


check_mem_header(0,"malloc-1");
	if ( dumy_mem == 0 ) {
		dumy_mem = S_MALLOC(1000);
		if ( dumy_mem == 0 )
			er_panic("no memory!!");
	}
	if ( (m = S_MALLOC(size+SIZEOF_MEM)) == 0 ) {
		S_FREE(dumy_mem);
		er_panic("no memory!!");
	}
check_mem_header(0,"malloc-2");
	m->m_size = size;
	m->m_flags = 0;
	m->file = __file;
	m->line = __line;
	ret = MEM2PT(m);
	pt = ret;
	for ( i = 0 ; i < size ; i ++ )
		*pt++ = 0;
check_mem_header(0,"malloc-3");
	lock_mem();
check_mem_header(0,"malloc-4");
	INSERT_MEM(&mem_header,m);
check_mem_header(0,"malloc-5");
	malloc_count += size;
check_mem_header(0,"malloc-6");
	gc_set_nl_non_check(ret,gc_func);
check_mem_header(0,"malloc-7");
	unlock_mem();
check_mem_header(0,"malloc-8");
	return ret;
}


void
mfree(void * f)
{
MEM * m;
	lock_mem(); 
	m = PT2MEM(f);
	malloc_count -= m->m_size;
	DELETE_MEM(m);
	gc_delete_nl(f);
	unlock_mem();
	S_FREE(m);
}

void *
mrealloc(void * pt,unsigned long size,void (*gc_func)())
{
void * ret;
MEM * m, * new;
//int i;
int s;
char * src, * dest;

	lock_mem();
	if ( pt == 0 )
		er_panic("mrealloc(1)");
	m = PT2MEM(pt);
	if ( m->m_size == size ) {
		unlock_mem();
		return pt;
	}
	new = S_MALLOC(size + SIZEOF_MEM);
	if ( new == 0 )
		er_panic("no memmory");
	new->m_size = size;
	new->m_flags = 0;
	new->file = m->file;
	new->line = m->line;
	ret = MEM2PT(new);
	if ( m->m_size < size )
		s = m->m_size;
	else 	s = size;
	src = pt;
	dest = ret;
/*
	for ( i = 0 ; i < s ; i ++ )
		*dest ++ = *src ++;
	for ( ; i < size ; i ++ )
		*dest ++ = 0;
*/
	memcpy(dest,src,s);
	memset(dest+s,0,size-s);
	m->m_flags |= M_STACK;
	INSERT_MEM(&mem_header,new);
	malloc_count += size;
	gc_set_nl_non_check(ret,gc_func);
	unlock_mem();
	return ret;
}


extern void *test_ptr;


int
get_ms_key(MEM * m)
{
	return m->line % MS_NODE_HASH_SIZE;
}

void
insert_mem_statistics(MEM_STATISTICS * ms,MEM * m)
{
int key;
MS_NODE * n;
	key = get_ms_key(m);
	for ( n = ms->hash[key] ; n ; n = n->next ) {
		if ( n->line != m->line )
			continue;
		if ( strcmp(n->file,m->file) )
			continue;
		n->total_size += m->m_size;
		return;
	}
	n = S_MALLOC(sizeof(*n));
	n->total_size = m->m_size;
	n->file = m->file;
	n->line = m->line;
	n->next = ms->hash[key];
	ms->hash[key] = n;
}

int
sort_trace(AVT_NODE * a,AVT_NODE ** ret)
{
	*ret = a;
	return 1;
}

int
sort_cmp(MS_NODE * n1,MS_NODE * n2)
{
	if ( n1->total_size < n2->total_size )
		return -1;
	if ( n1->total_size > n2->total_size )
		return 1;
	if ( ((long)n1) < ((long)n2) )
		return -1;
	if ( ((long)n1) > ((long)n2) )
		return 1;
	return 0;
}

void
sort_ms(MEM_STATISTICS * ms)
{
MS_NODE * n;
int i;
AVT_NODE * at;
	for ( i = 0 ; i < MS_NODE_HASH_SIZE ; i ++ )
		for ( n = ms->hash[i] ; n ; n = n->next ) {
			if ( ms->a_ptr < MS_NODE_HASH_SIZE ) {
				at = &ms->a[ms->a_ptr++];
			}
			else {
				avt_trace_from_small(ms->root,sort_trace,&at);
				at = avt_delete(&ms->root,at->data,sort_cmp);
			}
			at->data = n;
			avt_insert(&ms->root,at,sort_cmp);
		}
}

void
free_mem_statistics(MEM_STATISTICS * ms)
{
int i;
MS_NODE * n;
	for ( i = 0 ; i < MS_NODE_HASH_SIZE ; i ++ ) {
		for ( ; ms->hash[i] ; ) {
			n = ms->hash[i];
			ms->hash[i] = n->next;
			S_FREE(n);
		}
	}
}

int
_mem_gc(MEM_STATISTICS * ms)
{
MEM * m,* m1;
//int i;
int size;
int ret;
	ret = 0;
	if ( ms )
		memset(ms,0,sizeof(*ms));
	for ( m = mem_header.m_next ; m != &mem_header ; ) {
		if ( m->m_flags ) {
			m->m_flags = 0;
			if ( ms )
				insert_mem_statistics(ms,m);
			m = m->m_next;
			continue;
		}
		m1 = m;
		m = m->m_next;
		DELETE_MEM(m1);
		size = m1->m_size;
//		for ( i = 0 ; i < size+sizeof(MEM) ; i ++ )
//			((char*)m1)[i] = 0x79;
//printf("mem_gc %p\n",m1);
		S_FREE(m1);
//printf("mem_gc %p END\n",m1);
		ret += size;
	}
	malloc_count = malloc_count-ret;
	if ( ms )
		sort_ms(ms);
	return ret;
}

int
mem_gc(MEM_STATISTICS * ms)
{
int ret;
	lock_mem();
	ret = _mem_gc(ms);
	unlock_mem();
	return ret;
}

char *
copy_mstr(char * str)
{
char * ret;
	ret = mmalloc(strlen(str)+1,gc_text);
	strcpy(ret,str);
	return ret;
}

void
gc_text(char* tx)
{
	if ( tx == 0 )
		return;
	if ( TEST_AND_SET(tx) )
		return;
}

#define TEST_CNT	2000

void
mem_test(char * msg)
{
int i,j;
char ** t;

	printf("%s...",msg);
	fflush(stdout);

	t = d_alloc(sizeof(char*)*TEST_CNT);
	for ( i = 0 ; i < TEST_CNT ; i ++ ) {
		t[i] = d_alloc(i+1);
		for ( j = 0 ; j < i+1 ; j ++ )
			t[i][j] = 0x77;
	}
	for ( i = 0 ; i < TEST_CNT ; i ++ )
		d_f_ree(t[i]);
	d_f_ree(t);

	printf("END\n");

}

